# -*- coding: utf-8 -*-

from .. import EloquentTestCase
from .. import mock

from ..utils import MockConnection, MockProcessor

from eloquent.support.collection import Collection


class SupportCollectionTestCase(EloquentTestCase):

    pass
